# span_linking_dataset#handle indirect links(remove or leave), handle nulls
# span_recognition_dataset#assign all links to one of small concept set (remove if not a child of one of them), handle nulls
# concept_linking_dataset#add all concept ids as a document attribute
# entity_linking_dataset#handle indirect links(remove or leave), merge_spans(leaves or root),handle nulls
# entity_recognition_dataset#assign all links to one of a small concept set (remove if not a child of one of them), merge_spans(leaves or root),handle nulls


# null handling will be different based on if task implementation is end-to-end or modular.
# e.g. entities can be extracted but not linked.
from abc import ABC, abstractmethod
from collections import defaultdict
import copy
import pathlib
from typing import Dict, Iterable, List, Literal, Union
from spacy.tokens import Span, Doc, SpanGroup, DocBin
from spacy import displacy
from crtl.lkb.lkb import Concept, LexicalKnowledgeBase
from crtl.s3_file_utils.s3_file_utils import cached_path
import warnings
# trendnet: LexicalKnowledgeBase = LexicalKnowledgeBase.from_gzipped_jsonl(
#     cached_path("s3://databrickstrenddata/crtl/lkb/trendnet.gzip")
# )
class DeriverComponent(ABC):
    def __call__(self, spans: List[Span]) -> List[Span]:
        raise NotImplementedError


class DeriverPipeline:
    def __init__(self, components: List[DeriverComponent]):
        self._components = components
        self._output_container_type = (
            "ENT"
            if SpanMerger in [type(component) for component in self._components]
            else "SPAN"
        )

    def __call__(self, docs: Iterable[Doc]) -> None:
        for doc in docs:
            spans = [span for span in doc.spans["sc"]]
            for component in self._components:
                spans = component(spans)
            self._set_spans_or_ents(doc, spans)

    def _set_spans_or_ents(self, doc: Doc, spans: List[Span]) -> None:
        new_spans = []
        for span in spans:
            if span.label_ == "":
                label = "ENTITY"
            else:
                label = span.label_
            new_spans.append(Span(doc,span.start,span.end,label=label,kb_id=span.kb_id_))
        if self._output_container_type == "ENT":
            doc.set_ents(new_spans)
            doc.spans.clear()
        elif self._output_container_type == "SPAN":
            group = SpanGroup(doc, name="sc", spans=new_spans)
            doc.spans["sc"] = group
        
        


class SpanMerger(DeriverComponent):
    def __init__(self, span_merging_strategy: Literal["NARROW", "BROAD"]):
        self._span_merging_strategy = span_merging_strategy

    @classmethod
    def _contains(self, span_1, span_2):
        return span_1.start <= span_2.start and span_1.end >= span_2.end

    def _filter(self, span_1, span_2):
        if self._span_merging_strategy == "NARROW":
            return self._contains(span_1, span_2)
        elif self._span_merging_strategy == "BROAD":
            return self._contains(span_2, span_1)

    def __call__(self, spans: List[Span]):
        filtered_spans = []
        for span_1 in spans:
            for span_2 in spans:
                if span_1 != span_2:
                    if self._filter(span_1, span_2):
                        break
            else:
                filtered_spans.append(span_1)
        return filtered_spans


class ConceptGeneralizer(DeriverComponent):
    def __init__(self, high_level_kb_ids: List[str], lkb: LexicalKnowledgeBase):
        self._generalize_kb_ids = high_level_kb_ids
        self._low_to_high_map = self._generate_low_to_high_mapping(
            high_level_kb_ids, lkb
        )
        self._lkb = lkb

    def __call__(self, spans: List[Span]) -> List[Span]:
        for span in spans:
            if self._low_to_high_map[span.kb_id_]:
                span.kb_id_ = self._low_to_high_map[span.kb_id_][0]
                span.label_ = self._lkb.cui_to_concept[span.kb_id_].canonical_name
            else:
                span.kb_id_ = "NIL"
                span.label_ = "NIL"
        return spans

    def _generate_low_to_high_mapping(
        self, high_level_kb_ids: List[Union[str, int]], lkb: LexicalKnowledgeBase
    ) -> Dict[Union[str, int], Union[str, int]]:
        # Function to collect the nodes in a breadth-first traversal
        low_to_high_map = defaultdict(lambda: [])
        for high_level_concept in [
            lkb.cui_to_concept[kb_id] for kb_id in high_level_kb_ids
        ]:
            low_to_high_map[high_level_concept.concept_id].append(
                high_level_concept.concept_id
            )
            descendants = self._get_descendants(high_level_concept, lkb)
            for descendant in descendants:
                low_to_high_map[descendant.concept_id].append(
                    high_level_concept.concept_id
                )
        return low_to_high_map

    @classmethod
    def _get_descendants(
        cls, concept: Concept, lkb: LexicalKnowledgeBase
    ) -> List[Concept]:
        descendants = []
        visited = set()
        queue = []
        queue.append(concept)
        visited.add(concept.concept_id)
        while queue:
            concept = queue.pop(0)
            for concept_relation in concept.related_concepts:
                child_concept = lkb.cui_to_concept[concept_relation.concept_id]
                if child_concept not in visited:
                    if concept_relation.relation in ["inverse_isa", "has_tradename"]:
                        queue.append(child_concept)
                        descendants.append(child_concept)
                    visited.add(concept)
        return descendants


class NullRemover(DeriverComponent):
    def __call__(self, spans: List[Span]) -> List[Span]:
        return [span for span in spans if span.kb_id_ != "NIL"]


class IndirectLinkNullifier(DeriverComponent):
    def __call__(self, spans: List[Span]) -> List[Span]:
        for span in spans:
            if span._.is_broad:
                span.kb_id_ = "NIL"
        return spans


class DatasetDeriver(ABC):  # interface to all the dataset derivers.
    @abstractmethod
    def __call__(self, docs: Iterable[Doc]) -> List[Doc]:
        raise NotImplementedError
    def _to_disk(self,  docs: Iterable[Doc],path) -> None:
        docbin = DocBin(docs=docs)
        docbin.to_disk(path)
    def create_dataset(
        self, docs: Iterable[Doc], path: Union[str, pathlib.Path]
    ) -> None:
        docs = self(docs)
        self._to_disk(docs, path)
    @abstractmethod
    def create_html(self,docs:Iterable[Doc]):
        raise NotImplementedError
    # @abstractmethod
    # def display(self,docs:Iterable[Doc])->HTML:
    # pass


class SpanBasedDatasetDeriver(DatasetDeriver):
    _deriver_pipeline: DeriverPipeline
    _entity_based: bool

    def __call__(self, docs: Iterable[Doc]) -> List[Doc]:
        self._deriver_pipeline(docs)
        if self._entity_based:
            docs = self._filter_out_invalid_docs(docs)
        return docs

    def create_dataset(
        self, docs: Iterable[Doc], path: Union[str, pathlib.Path]
    ) -> None:
        if type(path) == str:
            path = pathlib.Path(path)
        docs = self(docs)
        self._to_disk(docs, path)

    @staticmethod
    def _filter_out_invalid_docs(docs: Iterable[Doc]) -> List[Doc]:
        filtered_docs = []
        for doc in docs:
            for ent in doc.ents:
                if True in [
                    (tok.ent_iob_ in ("B", "I") and tok.is_space) for tok in ent
                ]:
                    break
            else:
                filtered_docs.append(doc)
        return filtered_docs

class LinkedSpanDeriver(SpanBasedDatasetDeriver):
    def __init__(self, remove_nulls: bool, nullify_indirect_links: bool,lkb:LexicalKnowledgeBase):
        self._entity_based = False
        self._lkb = lkb
        pipeline_components: List[DeriverComponent] = []
        if nullify_indirect_links:
            pipeline_components.append(IndirectLinkNullifier())
        if remove_nulls:
            pipeline_components.append(NullRemover())
        self._deriver_pipeline = DeriverPipeline(pipeline_components)
    def create_html(self,docs:Iterable[Doc]):
        print("AHHHHH")
        for doc in docs:
            spans:Iterable[Span] = doc.spans["sc"]
            new_spans = []
            for span in spans:
                if span.kb_id_ == "NIL":
                    label = "NIL"
                elif span.kb_id_ not in self._lkb.cui_to_concept:
                    warnings.warn("Found a CUI in dataset not found in the LKB.")
                    label = "NIL"
                else:
                    label = self._lkb.cui_to_concept[span.kb_id_].canonical_name
                new_spans.append(Span(doc,span.start,span.end,label=label,kb_id=span.kb_id_))
            doc.spans["sc"] = SpanGroup(doc,name="sc",spans=new_spans)
        return displacy.render([doc for doc in docs if doc], style="span", page=True)

class RecognizedSpanDeriver(SpanBasedDatasetDeriver):
    def __init__(
        self,
        remove_nulls: bool,
        high_level_kb_ids: list[str],
        lkb: LexicalKnowledgeBase,
    ):
        self._entity_based = False
        self._lkb = lkb
        pipeline_components: List[DeriverComponent] = []
        pipeline_components.append(ConceptGeneralizer(high_level_kb_ids, lkb))
        if remove_nulls:
            pipeline_components.append(NullRemover())
        self._deriver_pipeline = DeriverPipeline(pipeline_components)
    def create_html(self,docs:Iterable[Doc]):
        return displacy.render([doc for doc in docs if doc], style="span", page=True)


class LinkedEntityDeriver(SpanBasedDatasetDeriver):
    def __init__(
        self,
        nullify_indirect_links: bool,
        span_merging_strategy: Literal["NARROW", "BROAD"],
        remove_nulls: bool,
        lkb:LexicalKnowledgeBase
    ):  # Fixes ordering
        self._entity_based = True
        self._lkb = lkb
        pipeline_components: List[DeriverComponent] = []
        if nullify_indirect_links:
            pipeline_components.append(IndirectLinkNullifier())
        if remove_nulls:
            pipeline_components.append(NullRemover())
        pipeline_components.append(SpanMerger(span_merging_strategy))
        self._deriver_pipeline = DeriverPipeline(pipeline_components)
    def create_html(self,docs:Iterable[Doc]):
        adjusted_docs:List[Doc] = []
        for doc in docs:
            doc = copy.deepcopy(doc)
            spans:Iterable[Span] = doc.ents
            for span in spans:
                if span.kb_id_ == "NIL":
                    span.label_ = "NIL"
                elif span.kb_id_ not in self._lkb.cui_to_concept:
                    warnings.warn("Found a CUI in dataset not found in the LKB.")
                    span.label_ = "NIL"
                else:
                    span.label_ = self._lkb.cui_to_concept[span.kb_id_].canonical_name
            doc.set_ents(spans)
            adjusted_docs.append(doc)
        return displacy.render([doc for doc in adjusted_docs if doc], style="ent", page=True)

class RecognizedEntityDeriver(SpanBasedDatasetDeriver):
    def __init__(
        self,
        span_merging_strategy: Literal["NARROW", "BROAD"],
        high_level_kb_ids: List[str],
        remove_nulls: bool,
        lkb: LexicalKnowledgeBase,
    ):  # Fixes Ordering
        self._entity_based = True
        self._lkb = lkb
        pipeline_components: List[DeriverComponent] = []
        pipeline_components.append(ConceptGeneralizer(high_level_kb_ids, lkb))
        if remove_nulls:
            pipeline_components.append(NullRemover())
        pipeline_components.append(SpanMerger(span_merging_strategy))
        self._deriver_pipeline = DeriverPipeline(pipeline_components)
    @classmethod
    def create_html(cls,docs:Iterable[Doc]):
        return displacy.render([doc for doc in docs if doc], style="ent", page=True)

class TaggedConceptsDeriver:
    Doc.set_extension("kb_ids", default=[])

    def __init__(self, with_broad_links: bool):
        self._with_broad_links = with_broad_links

    def __call__(self, docs: Iterable[Doc]) -> List[Doc]:
        for doc in docs:
            for span in doc.spans["sc"]:
                if self._test_for_inclusion(span):
                    doc._.kb_ids.append(span.kb_id_)
        return docs

    def _test_for_inclusion(self, span):
        return span.kb_id_ != "NIL" and (self._with_broad_links or not span._.is_broad)
